#
# LUOS CLI tool

# *******************************************************************************
# Import packages
# *******************************************************************************
import argparse
import sys
import time
from multiprocessing import Process, Value
import json
from pyluos import Device
from pyluos.tools.discover import serial_discover
import numpy as np
import math
import crc8
import os
from ..io.serial_io import Serial
import serial
import struct

# *******************************************************************************
# Global Variables
# *******************************************************************************
BOOTLOADER_IDLE = 0
BOOTLOADER_START = 1
BOOTLOADER_STOP = 2
BOOTLOADER_READY = 3
BOOTLOADER_ERASE = 4
BOOTLOADER_BIN_CHUNK = 5
BOOTLOADER_BIN_END = 6
BOOTLOADER_CRC_TEST = 7
BOOTLOADER_APP_SAVED = 8
BOOTLOADER_RESET = 9
BOOTLOADER_READY_RESP = 16
BOOTLOADER_BIN_HEADER_RESP = 17
BOOTLOADER_ERASE_RESP = 18
BOOTLOADER_BIN_CHUNK_RESP = 19
BOOTLOADER_BIN_END_RESP = 20
BOOTLOADER_CRC_RESP = 21
BOOTLOADER_ERROR_SIZE = 32

FILEPATH = None
NB_SAMPLE_BY_FRAME_MAX = 127

RESP_TIMEOUT = 3
ERASE_TIMEOUT = 10
PROGRAM_TIMEOUT = 5

BOOTLOADER_SUCCESS = 0
BOOTLOADER_DETECT_ERROR = 1
BOOTLOADER_FLASH_ERROR = 2
BOOTLOADER_FLASH_BINARY_ERROR = 3
BOOTLOADER_FLASH_PORT_ERROR = 4
# *******************************************************************************
# Function
# *******************************************************************************

# *******************************************************************************
# @brief find routing table
# @param port connected to the luos gate
# @return an object containing routing table
# *******************************************************************************
def find_network(device):
    device._send({'detection': {}})
    startTime = time.time()
    state = device._poll_once()
    retry = 0
    while ('routing_table' not in state):
        if ('route_table' in state):
            print('version of luos not supported')
            return
        state = device._poll_once()
        if (time.time()-startTime > 1):
            retry = retry +1
            if retry > 5:
                # detection is not working
                sys.exit("Detection failed.")
            device._send({'detection': {}})
            startTime = time.time()

    return state

# *******************************************************************************
# @brief find nodes to program in the network
# @param target list, routing table
# @return a tuple with 2 lists : nodes to reboot and nodes to program
# *******************************************************************************
def create_target_list(args, state):
    bypass_node = False
    nodes_to_program = []
    nodes_to_reboot = []
    for node in state['routing_table']:
        # prevent programmation of node 1
        bypass_node = False
        if(node['node_id'] == 1):
            bypass_node = True
        # check if node is in target list
        if not (bypass_node):
            nodes_to_reboot.append(node['node_id'])
            for target in args.target:
                if(int(node['node_id']) == int(target)):
                    nodes_to_program.append(node['node_id'])
    for target in args.target:
        if not(int(target) in nodes_to_program):
            print ("**** Node " + target + " is not available and will be ignored. ****")

    return (nodes_to_reboot, nodes_to_program)

# *******************************************************************************
# @brief send commands
# @param command type
# @return None
# *******************************************************************************
def send_topic_command(device, topic, command, size = 0):
    # create a json file with the list of the nodes to program
    bootloader_cmd = {
        'bootloader': {
            'command': {
                'type': command,
                'topic': topic,
                'size': size
            },
        }
    }
    # send json command
    device._send(bootloader_cmd)

def send_node_command(device, node, topic, command, size = 0):
    # create a json file with the list of the nodes to program
    bootloader_cmd = {
        'bootloader': {
            'command': {
                'type': command,
                'node': node,
                'topic': topic,
                'size': size
            },
        }
    }
    # send json command
    device._send(bootloader_cmd)

# *******************************************************************************
# @brief get binary size
# @param
# @return binary size
# *******************************************************************************
def get_binary_size():
    # get number of bytes in binary file
    with open(FILEPATH, mode="rb") as f:
        nb_bytes = len(f.read())

    return nb_bytes

# *******************************************************************************
# @brief send erase command
# @param command type
# @return None
# *******************************************************************************
def send_ready_cmd(device, node, topic):
    return_value = True
    # send ready command to the node
    send_node_command(device, node, topic, BOOTLOADER_READY, get_binary_size())
    # wait ready response
    state = device._poll_once()
    init_time = time.time()
    while (time.time() - init_time <= RESP_TIMEOUT):
        if 'bootloader' in state:
            for response in state['bootloader']:
                if response['response'] == BOOTLOADER_ERROR_SIZE:
                    print(u"  ╰> Node n°", response['node'], "has not enough space in flash memory.")
                    # don't load binary if there is not enough place in flash memory
                    return_value = False
                else:
                    print(u"  ╰> Node n°", response['node'], "is ready.")
                    return_value = True
            break

        state = device._poll_once()
        return_value = False

    return return_value
# *******************************************************************************
# @brief waiting for erase response
# @param
# @return binary size
# *******************************************************************************
def waiting_erase():
    count = 0
    period = 0.4
    print("\r                        ", end='')
    print(u"\r  ╰> Erase flash        ", end='')
    while(1):
        time.sleep(period)
        if(count == 0):
            print("\r                        ", end='')
            print(u"\r  ╰> Erase flash .      ", end='')
            count += 1
            continue
        if(count == 1):
            print("\r                        ", end='')
            print(u"\r  ╰> Erase flash ..     ", end='')
            count += 1
            continue
        if(count == 2):
            print("\r                        ", end='')
            print(u"\r  ╰> Erase flash ...    ", end='')
            count += 1
            continue
        if(count == 3):
            print("\r                        ", end='')
            print(u"\r  ╰> Erase flash        ", end='')
            count = 0
            continue

# *******************************************************************************
# @brief send erase command
# @param command type
# @return None
# *******************************************************************************
def erase_flash(device, topic, nodes_to_program):
    return_value = True
    failed_nodes = []
    failed_nodes.extend(nodes_to_program)
    timeout = ERASE_TIMEOUT * len(nodes_to_program)

    # send erase command
    send_topic_command(device, topic, BOOTLOADER_ERASE)

    # display a progress bar
    waiting_bg = Process(target=waiting_erase)
    waiting_bg.start()

    # pull serial data
    state = device._poll_once()
    # initialize the timer that counts until node number * response time
    init_time = time.time()
    # check if all messages are received
    while len(failed_nodes):
        # timeout for exiting loop in case of fails
        if(time.time() - init_time > timeout):
            return_value = False
            break
        # check if it is a response message
        if 'bootloader' in state:  
            for response in state['bootloader']:
                if (response['response'] == BOOTLOADER_ERASE_RESP):
                    # this node responded, delete it from the failed nodes list
                    if response['node'] in failed_nodes:
                        timeout -= ERASE_TIMEOUT
                        failed_nodes.remove(response['node'])
                        print(u"\r\n  ╰> Flash memory of node ", response['node'], " erased.")
        state = device._poll_once()

    # retry sending failed messages
    for node in failed_nodes:
        send_node_command(device, node, topic, BOOTLOADER_ERASE)
        print(u"\r\n  ╰> Retry erase memory of node ", node)
        init_time = time.time()
        state = device._poll_once()
        while len(failed_nodes):
            if(time.time() - init_time > ERASE_TIMEOUT):
                return_value = False
                break

            # check if it is a response message
            if 'bootloader' in state:  
                for response in state['bootloader']:
                    if (response['response'] == BOOTLOADER_ERASE_RESP):
                    # this node responded, delete it from the failed nodes list
                        if response['node'] in failed_nodes:
                            failed_nodes.remove(response['node'])
                            print(u"\r\n  ╰> Flash memory of node ", response['node'], " erased.")
            state = device._poll_once()

    if not len(failed_nodes):
        return_value = True

    waiting_bg.terminate()
    return return_value, failed_nodes

# *******************************************************************************
# @brief get binary size
# @param
# @return binary size
# *******************************************************************************
def loading_bar(loading_progress):
    period = 0.2
    while(1):
        time.sleep(period)
        print(u"\r  ╰> loading : {} %".format(loading_progress.value), end='')

# *******************************************************************************
# @brief send the binary file to the node
# @param command type
# @return None
# *******************************************************************************
def send_binary_data(device, topic, nodes_to_program):
    loading_state = True
    failed_nodes = []
    prev_fails = []
    # compute total number of bytes to send
    with open(FILEPATH, mode="rb") as f:
        nb_bytes = len(f.read())
    # compute total number of frames to send
    nb_frames = math.ceil(nb_bytes / NB_SAMPLE_BY_FRAME_MAX)

    # display a progress bar to inform user
    loading_progress = Value('f', 0.0)
    loading_bar_bg = Process(target=loading_bar, args=(loading_progress,))
    loading_bar_bg.start()

    # send each frame to the network
    file_offset = 0
    for frame_index in range(nb_frames):
        if (frame_index == (nb_frames-1)):
            # last frame, compute size
            frame_size = nb_bytes - (nb_frames-1)*NB_SAMPLE_BY_FRAME_MAX
        else:
            frame_size = NB_SAMPLE_BY_FRAME_MAX

        # send the current frame
        loading_state, failed_nodes = send_frame_from_binary(device, topic, frame_size, file_offset, nodes_to_program)
        if not loading_state:
            print(u"\r  ╰> loading of node: ", failed_nodes, "FAILED %")
            for fail in failed_nodes:
                nodes_to_program.remove(fail)
            prev_fails.extend(failed_nodes)
            loading_state = True
        # update cursor position in the binary file
        file_offset += frame_size
        # update loading progress
        loading_progress.value = math.trunc(frame_index / nb_frames * 100)
        if not len(nodes_to_program):
            break;

    # kill the progress bar at the end of the loading
    loading_bar_bg.terminate()
    if loading_state:
        print(u"\r  ╰> loading : 100.0 %")
    if len(prev_fails):
        loading_state = False
    return loading_state, prev_fails

# *******************************************************************************
# @brief open binary file and send a frame
# @param
# @return None
# *******************************************************************************
def send_frame_from_binary(device, topic, frame_size, file_offset, nodes_to_program):
    return_value = True
    failed_nodes = []
    failed_nodes.extend(nodes_to_program)
    timeout = PROGRAM_TIMEOUT * len(nodes_to_program)

    with open(FILEPATH, mode="rb") as f:
        # put the cursor at the beginning of the file
        f.seek(file_offset)
        # read binary data
        data_bytes = f.read(1)
        for sample in range(frame_size-1):
            data_bytes = data_bytes + f.read(1)

    send_data(device, topic, BOOTLOADER_BIN_CHUNK, frame_size, data_bytes)
    # pull serial data
    state = device._poll_once()
    # wait nodes response
    init_time = time.time()
    while len(failed_nodes):
        # check for timeout of nodes
        if(time.time() - init_time > timeout):
            return_value = False
            break
        # check if it is a response message
        if 'bootloader' in state:
            for response in state['bootloader']:
                if (response['response'] == BOOTLOADER_BIN_CHUNK_RESP):
                    # the node responsed, remove it for fails list
                    if response['node'] in failed_nodes:
                        timeout -= PROGRAM_TIMEOUT
                        failed_nodes.remove(response['node'])
        # wait for next message
        state = device._poll_once()

    for node in failed_nodes:
        # retry sending failed messages
        send_data_node(device, node, BOOTLOADER_BIN_CHUNK, frame_size, data_bytes)
        state = device._poll_once()
        print(u"\r\n  ╰> Retry sending binary message to node ", node)
        init_time = time.time()
        while len(failed_nodes):
            # check for timeout of nodes
            if(time.time() - init_time > PROGRAM_TIMEOUT):
                return_value = False
                break

            # check if it is a response message
            if 'bootloader' in state:
                for response in state['bootloader']:
                    if (response['response'] == BOOTLOADER_BIN_CHUNK_RESP):
                        # the node responsed, remove it for fails list
                        if response['node'] in failed_nodes:
                            failed_nodes.remove(response['node'])
            # wait for next message
            state = device._poll_once()

    if not len(failed_nodes):
        return_value = True

    return return_value, failed_nodes

# *******************************************************************************
# @brief send binary data with a header
# @param
# @return None
# *******************************************************************************
def send_data(device, topic, command, size, data):
    # create a json file with the list of the nodes to program
    bootloader_cmd = {
        'bootloader': {
            'command': {
                'size': [size],
                'type': command,
                'topic': topic,
            },
        }
    }
    # send json command
    device._write(json.dumps(bootloader_cmd).encode() + '\n'.encode() + data)

# *******************************************************************************
# @brief send binary data with a header to a specific node
# @param
# @return None
# *******************************************************************************
def send_data_node(device, node, command, size, data):
    # create a json file with the list of the nodes to program
    bootloader_cmd = {
        'bootloader': {
            'command': {
                'size': [size],
                'type': command,
                'node': node,
            },
        }
    }
    # send json command
    device._write(json.dumps(bootloader_cmd).encode() + '\n'.encode() + data)

# *******************************************************************************
# @brief send the binary end command
# @param
# @return
# *******************************************************************************
def send_binary_end(device, topic, nodes_to_program):
    return_value = True
    failed_nodes = []
    failed_nodes.extend(nodes_to_program)
    timeout = RESP_TIMEOUT * len(nodes_to_program)
    # send command
    send_topic_command(device, topic, BOOTLOADER_BIN_END)
    # poll serial data
    state = device._poll_once()
    # wait bin_end response
    init_time = time.time()
    while len(failed_nodes) > 0:
        # check if we exit with timeout
        if(time.time() - init_time > timeout):
            return_value = False
            break
        if 'bootloader' in state:
            for response in state['bootloader']:
                # check each node response
                if (response['response'] == BOOTLOADER_BIN_END_RESP):
                    print(u"  ╰> Node", response['node'], "acknowledge received, loading is complete.")
                    # remove node from fails list
                    if response['node'] in failed_nodes:
                        timeout -= RESP_TIMEOUT
                        failed_nodes.remove(response['node'])
        state = device._poll_once()

    for node in failed_nodes:
        # retry sending failed messages
        send_node_command(device, node, topic, BOOTLOADER_BIN_END)
        print(u"\r\n  ╰> Retry sending end message to node ", node)
        state = device._poll_once()
        init_time = time.time()
        while len(failed_nodes):
            # check for timeout of nodes
            if(time.time() - init_time > RESP_TIMEOUT):
                return_value = False
                break

            # check if it is a response message
            if 'bootloader' in state:
                for response in state['bootloader']:
                    if (response['response'] == BOOTLOADER_BIN_END_RESP):
                        # the node responsed, remove it for fails list
                        if response['node'] in failed_nodes:
                            failed_nodes.remove(response['node'])
            # wait for next message
            state = device._poll_once()

    if not len(failed_nodes):
        return_value = True

    return return_value, failed_nodes

# *******************************************************************************
# @brief compute binary crc
# @param
# @return None
# *******************************************************************************
def compute_crc():
    # create crc8 function object
    hash = crc8.crc8()
    # get number of bytes in binary file
    with open(FILEPATH, mode="rb") as f:
        nb_bytes = len(f.read())

    with open(FILEPATH, mode="rb") as f:
        for bytes in range(nb_bytes):
            data = f.read(1)
            hash.update(data)
            crc = hash.digest()

    return crc

# *******************************************************************************
# @brief send the binary end command
# @param
# @return
# *******************************************************************************
def check_crc(device, topic, nodes_to_program):
    return_value = True
    failed_nodes = []
    failed_nodes.extend(nodes_to_program)
    timeout = RESP_TIMEOUT * len(nodes_to_program)

    # send crc command
    send_topic_command(device, topic, BOOTLOADER_CRC_TEST)
    
    state = device._poll_once()
    # wait bin_end response
    init_time = time.time()
    while len(failed_nodes):
        # check for timeout exit
        if(time.time() - init_time > timeout):
            return_value = False
            break
        # check the response
        if 'bootloader' in state:
            for response in state['bootloader']:
                if (response['response'] == BOOTLOADER_CRC_RESP):
                    source_crc = int.from_bytes(compute_crc(), byteorder='big')
                    node_crc = response['crc_value']
                    node_id = response['node']
                        # crc properly received 
                    if (source_crc == node_crc):
                        print(u"  ╰> CRC test for node", node_id, " : OK.")
                        if node_id in failed_nodes:
                            timeout -= RESP_TIMEOUT
                            failed_nodes.remove(node_id)
                    else:
                        # not a good crc
                        print(u"  ╰> CRC test for node", node_id, ": NOK.")
                        print(u"  ╰> waited :", hex(source_crc), ", received :", hex(node_crc))
                        return_value = False
        state = device._poll_once()

    for node in failed_nodes:
        # retry sending failed messages
        send_node_command(device, node, topic, BOOTLOADER_CRC_TEST)
        print(u"\r\n  ╰> Retry sending crc demand to node ", node)
        state = device._poll_once()
        init_time = time.time()
        while len(failed_nodes):
            # check for timeout of nodes
            if(time.time() - init_time > RESP_TIMEOUT):
                return_value = False
                break

            # check if it is a response message
            if 'bootloader' in state:
                for response in state['bootloader']:
                    if (response['response'] == BOOTLOADER_CRC_RESP):
                        source_crc = int.from_bytes(compute_crc(), byteorder='big')
                        node_crc = response['crc_value']
                        node_id = response['node']
                            # crc properly received 
                        if (source_crc == node_crc):
                            print(u"  ╰> CRC test for node", node_id, " : OK.")
                            if node_id in failed_nodes:
                                timeout -= RESP_TIMEOUT
                                failed_nodes.remove(node_id)
                        else:
                            # not a good crc
                            print(u"  ╰> CRC test for node", node_id, ": NOK.")
                            print(u"  ╰> waited :", hex(source_crc), ", received :", hex(node_crc))
                            return_value = False
            state = device._poll_once()

    if not len(failed_nodes):
        return_value = True

    return return_value, failed_nodes

# *******************************************************************************
# @brief reboot all nodes in application mode
# @param
# @return
# *******************************************************************************
def reboot_network(device, topic, nodes_to_reboot):
    for node in nodes_to_reboot:
                send_node_command(device, node, topic, BOOTLOADER_STOP)
                # delay to let gate send commands
                time.sleep(0.01)
# *******************************************************************************
# @brief command used to flash luos nodes
# @param flash function arguments : -g, -t, -b
# @return None
# *******************************************************************************
def luos_flash(args):
    print('Luos flash subcommand with parameters :')
    print('\t--baudrate : ', args.baudrate)
    print('\t--gate : ', args.gate)
    print('\t--target : ', args.target)
    print('\t--binary : ', args.binary)
    print('\t--port : ', args.port)

    topic = 1

    if not (args.port):
        try:
            args.port= serial_discover(os.getenv('LUOS_BAUDRATE', args.baudrate))[0]
        except:
            sys.exit()
            return

    # state used to check each step
    machine_state = True
    # list of all the nodes that may fail in each step
    total_fails = []
    # update firmware path
    global FILEPATH
    FILEPATH = args.binary
    try:
        f = open(FILEPATH, mode="rb")
    except IOError:
        sys.exit("Cannot open :", FILEPATH)
        return BOOTLOADER_FLASH_BINARY_ERROR
    else:
        f.close()

    # init device
    device = Device(args.port, baudrate=os.getenv('LUOS_BAUDRATE', args.baudrate), background_task=False)

    # find routing table
    state = find_network(device)

    # searching nodes to program in network
    (nodes_to_reboot, nodes_to_program) = create_target_list(args, state)

    # check if we have available node to program
    if not nodes_to_program:
        sys.exit("No target found :\n" + str(device.nodes))

    # reboot all nodes in bootloader mode
    print("** Reboot all nodes in bootloader mode **")
    for node in nodes_to_reboot:
        send_node_command(device, node, topic, BOOTLOADER_START)
        # delay to let gate send commands
        time.sleep(0.01)

    # wait before next step
    time.sleep(0.1)

    # find routing table in boot mode
    # its necessary to give ids to bootloader services
    state = find_network(device)

    # wait before next step
    time.sleep(0.4)

    print("\n** Programming nodes **")
    
    # go to header state if node is ready
    for node in nodes_to_program:
        print("--> Check if node", node, " is ready.")
        machine_state = send_ready_cmd(device, node, topic)
        if not machine_state:
            total_fails.append(node)
            machine_state = True
            print("Node ", node, "failed to load!")
        time.sleep(0.01)

    for node in total_fails:
        nodes_to_program.remove(node)

    # erase node flash memory
    print("--> Erase flash memory.")
    machine_state, failed_nodes = erase_flash(device, topic, nodes_to_program)
    if not machine_state:
        for fail in failed_nodes:
            nodes_to_program.remove(fail)
        total_fails.extend(failed_nodes)
        machine_state = True
        print("Erase flash of node: ", failed_nodes, "failed!")

    # send binary data
    print("--> Send binary data.")
    machine_state, failed_nodes = send_binary_data(device, topic, nodes_to_program)
    if not machine_state:
        total_fails.extend(failed_nodes)
        machine_state = True
        print("Flash of node: ", failed_nodes, "failed!")
    
    # inform the node of the end of the loading
    print("--> Programmation finished, waiting for acknowledge.")
    machine_state, failed_nodes = send_binary_end(device, topic, nodes_to_program)
    if not machine_state:
        for fail in failed_nodes:
            nodes_to_program.remove(fail)
        total_fails.extend(failed_nodes)
        machine_state = True
        print("ACK of node: ", failed_nodes, "failed!")

    # Ask the node to send binary crc
    print("--> Check binary CRC.")
    machine_state, failed_nodes = check_crc(device, topic, nodes_to_program)
    if not machine_state:
        for fail in failed_nodes:
            nodes_to_program.remove(fail)
        total_fails.extend(failed_nodes)
        machine_state = True
        print("ACK of node: ", failed_nodes, "failed!")

    # Say to the bootloader that the integrity
    # of the app saved in flash has been verified
    send_topic_command(device, topic, BOOTLOADER_APP_SAVED)

    # wait before next step
    time.sleep(0.1)
    # reboot all nodes in application mode
    if len(total_fails) == 0:
        print("** Reboot all nodes in application mode **")
        reboot_network(device, topic, nodes_to_reboot)
        device.close()
        return BOOTLOADER_SUCCESS
    else:
        device.close()
        print("Load of nodes: ", total_fails, " failed, please reboot and retry.")
        return BOOTLOADER_FLASH_ERROR

# *******************************************************************************
# @brief command used to detect network
# @param detect function arguments : -p
# @return None
# *******************************************************************************
def luos_detect(args):
    print('Luos detect subcommand on port : ', args.port)
    print('\tLuos detect subcommand at baudrate : ', args.baudrate)

    if not (args.port):
        try:
            args.port= serial_discover(os.getenv('LUOS_BAUDRATE', args.baudrate))[0]
        except:
            sys.exit()
            return

    # detect network
    device = Device(args.port, baudrate=os.getenv('LUOS_BAUDRATE', args.baudrate))
    # print network to user
    print(device.nodes)
    device.close()

    return BOOTLOADER_SUCCESS

# *******************************************************************************
# @brief command used to force reset a node in bootloader mode
# @param detect function arguments : -p
# @return None
# *******************************************************************************
def luos_reset(args):
    print('Luos discover subcommand on port : ', args.port)
    print('\tLuos discover subcommand at baudrate : ', args.baudrate)

    if not (args.port):
        try:
            args.port= serial_discover(os.getenv('LUOS_BAUDRATE', args.baudrate))[0]
        except:
            sys.exit()
            return


    # send rescue command
    print('Send reset command.')
    port = serial.Serial(args.port, os.getenv('LUOS_BAUDRATE', args.baudrate), timeout=0.05)
    rst_cmd = {
        'bootloader': {
            'command': {
                'type': BOOTLOADER_RESET,
                'node': 0,
                'size': 0
            },
        }
    }
    s = json.dumps(rst_cmd).encode()
    port.write(b'\x7E' + struct.pack('<H', len(s)) + s + b'\x81')
    port.close()

    # detect network
    device = Device(args.port, baudrate=os.getenv('LUOS_BAUDRATE', args.baudrate), background_task=False)

    print(device.nodes)
    device.close()

    return BOOTLOADER_SUCCESS

# *******************************************************************************
# @brief command used to detect network
# @param detect function arguments : -p
# @return None
# *******************************************************************************
def luos_options():
    parser = argparse.ArgumentParser(
        description='Luos command line interface')

    # create subcommands
    subparsers = parser.add_subparsers()

    # declare "flash" subcommand
    flash_parser = subparsers.add_parser('flash',
                                         help='tool to program luos nodes')
    flash_parser.add_argument('port',
                              help='port used to detect network',
                              nargs='?')
    flash_parser.add_argument('-g', '--gate',
                              help='id of the gate used to access the luos network')
    flash_parser.add_argument('-b', '--binary',
                              help='path to the binary file to flash')
    flash_parser.add_argument('-t', '--target',
                              help='target node to flash',
                              default=['2'],
                              nargs='*')
    flash_parser.add_argument('--baudrate',
                              help='Choose pyluos serial baudrate default value = 1000000',
                              default=1000000)
    flash_parser.set_defaults(func=luos_flash)

    # declare "detect" subcommand
    detect_parser = subparsers.add_parser('detect',
                                          help='tool to detect luos network')
    detect_parser.add_argument('port', help='port used to detect network',
                              nargs='?')
    detect_parser.add_argument('--baudrate',
                               help='Choose pyluos serial baudrate default value = 1000000',
                               default=1000000)
    detect_parser.set_defaults(func=luos_detect)

    # declare "rescue" subcommand
    rescue_parser = subparsers.add_parser('reset',
                                          help='tool to reset one or multiple blocked nodes in rescue mode')
    rescue_parser.add_argument('port', help='port used to access to the network',
                              nargs='?')
    rescue_parser.add_argument('--baudrate',
                               help='Choose pyluos serial baudrate default value = 1000000',
                               default=1000000)
    rescue_parser.set_defaults(func=luos_reset)

    return parser

# *******************************************************************************
# @brief parse arguments and launch subcommand function
# @param None
# @return None
# *******************************************************************************
def main():
    # declare options of the CLI
    parser = luos_options()

    # parse options
    args = parser.parse_args()

    # execute CLI subcommand
    return args.func(args)

if __name__ == '__main__':
    sys.exit(main())



from __future__ import division

import os
import json
import time
import serial as _serial
import platform
import sys
import struct
if sys.version_info >= (3, 0):
    import queue
else:
    import Queue as queue


from threading import Event, Thread

from serial.tools.list_ports import comports

from . import IOHandler

try:
    JSONDecodeError = json.decoder.JSONDecodeError
except AttributeError:
    JSONDecodeError = ValueError

class Serial(IOHandler):
    poll_frequency = 200
    period = 1 / poll_frequency
    message_rate = 12000 # Max messages per seconds

    @classmethod
    def available_hosts(cls):
        devices = comports(include_links=True)

        return [d.device for d in devices]

    @classmethod
    def is_host_compatible(cls, host):
        return host in cls.available_hosts()

    def __init__(self, host, baudrate=None):
        if baudrate is None:
            baudrate = os.getenv('LUOS_BAUDRATE', 1000000)

        self._serial = _serial.Serial(host, baudrate)
        self._serial.flush()

        self._msg = queue.Queue(int((self.message_rate / self.period) / 1000))
        self._running = True

        self._poll_loop = Thread(target=self._poll)
        self._poll_loop.daemon = True
        self._poll_loop.start()

    def is_ready(self):
        if self._serial.in_waiting == 0:
            return False

        try:
            self.read()
            return True
        except (UnicodeDecodeError, JSONDecodeError):
            return False

    def recv(self):
        try:
            data = self._msg.get(True, 1)
        except queue.Empty:
            data = None
        return data

    def write(self, data):
        self._serial.write(b'\x7E' + struct.pack('<H', len(data)) + data + b'\x81')

    def close(self):
        self._running = False
        self._poll_loop.join()

        self._serial.close()

    def _poll(self):
        def extract_line(s):
            # Find a serial header
            H = s.find(b'\x7E')
            if H == -1:
                # No header found
                return b'', s[0:0]
            else:
                # Header found, get size
                try:
                    size = struct.unpack('<H', s[H+1:H+3])[0]
                except:
                    # size not completely received
                    return b'', s[H:]
                if (size == 0) or (size > 20000):
                    # Bad header
                    # Remove the header and do it again
                    return extract_line(s[H+1:])
                else:
                    # Size seems ok
                    # Check if we receive the entire data
                    if len(s[H+3:]) < size+1:
                        # We don't have the entire data
                        return b'', s
                    else:
                        # We have the complete data
                        # Check the footer
                        data_start = H+3
                        data_end = data_start + size
                        if (s[data_end] != ord(b'\x81')):
                            # The footer is not ok, this mean we don't have a good header
                            # Remove the header and do it again
                            return extract_line(s[H+1:])
                        else:
                            # Footer is ok
                            return s[data_start:data_end], s[data_end + 1:]

        buff = b''

        while self._running:
            to_read = self._serial.in_waiting

            if to_read == 0:
                time.sleep(self.period)
                continue

            s = self._serial.read(to_read)
            buff = buff + s

            while self._running:
                line, buff = extract_line(buff)
                if not len(line):
                    break
                if self._msg.full():
                    print("Warning: Serial message queue is full. Some datas could be lost")
                    self._msg.get()
                self._msg.put(line)



import os
import time

from ..io.serial_io import Serial
import serial
import struct
import argparse

def serial_ports():
    return Serial.available_hosts()

def serial_discover(baudrate=1000000):
    serial_hosts = serial_ports()
    available_serial = []
    print("Searching for a gate available")
    for serial_host in serial_hosts:
        print("Search a Gate on port " + str(serial_host))
        try:
            port = serial.Serial(serial_host, baudrate, timeout=0.2)
            time.sleep(0.1)
        except:
            continue

        if port is not None:
            s = b'{}'
            port.write(b'\x7E' + struct.pack('<H', len(s)) + s + b'\x81')
            time.sleep(0.01)
            port.readline()
            port.flush()
            time.sleep(0.01)
            s = b'{\"discover\": {}}'
            port.write(b'\x7E' + struct.pack('<H', len(s)) + s + b'\x81')
            state = port.readline()
            gateResponse = False
            if 'gate'.encode() in state:
                gateResponse = True
            elif len(state):
                # if many other messages are received, drop them and retry reception
                for read_retry in range(1000):
                    state = port.readline()
                    if 'gate'.encode() in state:
                        gateResponse = True
                        break
            if gateResponse:
                available_serial.append(serial_host)
            port.reset_output_buffer()
            port.close()

    if available_serial :
        return available_serial
    else:
        print("... No gate detected")
        return []

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--baudrate", action="store",
                        help="Choose pyluos serial baudrate default value = 1000000",
                        default=1000000)

    args = parser.parse_args()

    gate_list = serial_discover(os.getenv('LUOS_BAUDRATE', args.baudrate))

    if gate_list:
        print("Available Luos gate on port : " + str(gate_list))
    else:
        print("No gate detected")


if __name__ == '__main__':
    main()



from .service import Service
import time

class Void(Service):

    def __init__(self, id, alias, device):
        Service.__init__(self, 'Void', id, alias, device)

    def _update(self, new_state):
        Service._update(self, new_state)

    def dxl_detect(self):
        self._push_value('reinit', 0)
        print ("To get new detected Dxl motors usable on pyluos you should recreate your Pyluos object.")

    def _factory_reset(self):
        new_val = [0xFF, 0]
        self._push_value('register', new_val)

    def factory_reset(self):
        self._factory_reset()
        print("Motor reseted => baudrate : 1000000, ID : same")
        print("you should start 'dxl_detect()' command and recreate your Pyluos object.")

    def retrieve_dxl(self):
        time.sleep(0.5)
        self.dxl_detect()
        print("Motor reseted => baudrate : 1000000, ID : same")
        print("recreate your Pyluos object.")



# -*- coding: utf-8 -*-
import os
import sys
import json
import time
import uuid
import logging
import requests
import threading
import logging.config
import numpy as np

from datetime import datetime
from collections import defaultdict

from .io import discover_hosts, io_from_host, Ws
from .services import name2mod

from anytree import AnyNode, RenderTree, DoubleStyle


def run_from_unittest():
    return 'unittest' in sys.services

class contList(list):
    def __repr__(self):
        s = '-------------------------------------------------\n'
        s += '{:<20s}{:<20s}{:<5s}\n'.format("Type", "Alias", "ID")
        s += '-------------------------------------------------\n'
        for elem in self:
            s += '{:<20s}{:<20s}{:<5d}\n'.format(elem.type, elem.alias, elem.id)
        return s

class nodeList(list):
    def __repr__(self):
        # Display the topology
        s = ''
        prefill = ''
        prechild = False
        for pre, fill, node in RenderTree(self[0], style=DoubleStyle()):
            child = []
            if (node.parent == None):
                branch = "  ┃  "
                for i,x in enumerate(node.port_table):
                    child.append(i)
            else:
                l_port_id = '?'
                for i,x in enumerate(node.parent.port_table):
                    if (x == node.id):
                        l_port_id = str(i)
                r_port_id = node.port_table.index(min(node.port_table))
                for i,x in enumerate(node.port_table):
                    if ((i != r_port_id) and (x != 65535)):
                        child.append(i)
                branch = str(l_port_id) + ">┃" + str(r_port_id) + " "
            prefill = (prefill[:len(fill)]) if len(prefill) > len(fill) else prefill
            s +='{:<{fillsize}s}'.format(prefill, fillsize=len(fill))
            if (prechild == True):
                position = -4
                s = s[:position] + '║' + s[position+1:]
            s += "  ┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓\n"
            tmpstr = "%s╭node %s" % (branch, node.id)
            s += pre + '{:^10s}'.format(tmpstr)
            if (node.certified == True):
                s += '{:^41s}'.format("Certified") + "┃\n"
            else:
                s += '{:^41s}'.format("/!\\ Not certified") + "┃\n"
            s += fill + "  ┃  │  " + '{:<20s}{:<20s}{:<5s}'.format("Type", "Alias", "ID")+ "┃\n"
            for y,elem in enumerate(node.services):
                if (y == (len(node.services)-1)):
                    s += fill + "  ┃  ╰> " + '{:<20s}{:<20s}{:<5d}'.format(elem.type, elem.alias, elem.id)+ "┃\n"
                else:
                    s += fill + "  ┃  ├> " + '{:<20s}{:<20s}{:<5d}'.format(elem.type, elem.alias, elem.id) + "┃\n"
            if (not child):
                s += fill + " >┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛\n"
                prechild = False
            else:
                s += fill + "╔>┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛\n"
                prechild = True
            prefill = fill
        return s


class Device(object):
    _heartbeat_timeout = 5  # in sec.
    _max_alias_length = 15
    _base_log_conf = os.path.join(os.path.dirname(__file__),
                                  'logging_conf.json')
    _freedomLink = None

    def __init__(self, host,
                 IO=None,
                 log_conf=_base_log_conf,
                 test_mode=False,
                 background_task=True,
                 telemetry=True,
                 *args, **kwargs):
        if IO is not None:
            self._io = IO(host=host, *args, **kwargs)
        else:
            self._io = io_from_host(host=host,
                                    *args, **kwargs)

        if os.path.exists(log_conf):
            with open(log_conf) as f:
                config = json.load(f)
            logging.config.dictConfig(config)

        self.telemetry = telemetry
        self.logger = logging.getLogger(__name__)
        self.logger.info('Connected to "{}".'.format(host))

        self._send_lock = threading.Lock()
        self._cmd_lock = threading.Lock()

        # We force a first poll to setup our model.
        self._setup()
        self.logger.info('Device setup.')

        self._last_update = time.time()
        self._running = True
        self._pause = False

        if(background_task == True):
            # Setup both poll/push synchronization loops.
            self._poll_bg = threading.Thread(target=self._poll_and_up)
            self._poll_bg.daemon = True
            self._poll_bg.start()

    def close(self):
        self._running = False

        if hasattr(self, "_poll_bg"):
            self._poll_bg.join(timeout=2.0)

            if self._poll_bg.is_alive():
                # _poll_bg didn't terminate within the timeout
                print("Warning: device closed on timeout, background thread is still running.")
        self._io.close()

    def link_to_freedomrobotics(self):
        from .integration.freedomRobotics import FreedomLink
        self._freedomLink = FreedomLink(self)

    def pause(self):
        self._pause = True
        time.sleep(1)

    def play(self):
        self._pause = False

    def _setup(self):
        self.logger.info('Sending detection signal.')
        self._send({})
        time.sleep(0.01)
        self._send({'detection': {}})
        self.logger.info('Waiting for routing table...')
        startTime = time.time()
        state = self._poll_once()
        retry = 0
        while ('routing_table' not in state):
            if ('route_table' in state):
                self.logger.info("Watch out the Luos revision you are using on your board is too old to work with this revision on pyluos.\n Please consider updating Luos on your boards")
                return
            state = self._poll_once()
            if (time.time()-startTime > 1):
                retry = retry +1
                if retry > 5:
                    # detection is not working
                    sys.exit("Detection failed.")
                self._send({'detection': {}})
                startTime = time.time()
        # Create nodes
        self._services = []
        self._nodes = []
        for i, node in enumerate(state['routing_table']):
            if ('node_id' not in node):
                self.logger.info("Watch out the Luos revision you are using on your board is too old to work with this revision on pyluos.\n Please consider updating Luos on your boards")
            parent_elem = None
            # find a parent and create a link
            if (min(node["port_table"]) < node["services"][0]["id"]):
                parent_id = min(node["port_table"])
                for elem in self._nodes:
                    if (elem.id == parent_id):
                        parent_elem = elem
                        break;
            # create the node
            self._nodes.append(AnyNode(id=node["node_id"], certified=node["certified"], parent=parent_elem, port_table=node["port_table"]))

            filtered_services = contList([mod for mod in node["services"]
                                if 'type' in mod and mod['type'] in name2mod.keys()])
            # Create a list of services in the node
            self._nodes[i].services = [
                name2mod[mod['type']](id=mod['id'],
                                      alias=mod['alias'],
                                      device=self)
                for mod in filtered_services
                if 'type' in mod and 'id' in mod and 'alias' in mod
            ]
            # Create a list of services of the entire device
            self._services = self._services + self._nodes[i].services
            for mod in self._nodes[i].services:
                setattr(self, mod.alias, mod)

        self._cmd = defaultdict(lambda: defaultdict(lambda: None))
        self._cmd_data = []
        self._binary = []

        if (self.telemetry == True):
            from pyluos.version import version
            self.logger.info('Sending telemetry...')
            luos_telemetry = {"telemetry_type": "pyluos",
                  "mac": hex(uuid.getnode()),
                  "system": sys.platform,
                  "unix_time": int(time.time()),
                  "pyluos_rev": version,
                  "routing_table":state['routing_table']}
            try:
                r = requests.post("https://monorepo-services.vercel.app/api/telemetry",
                           data=luos_telemetry)
                if not r:
                    print("Telemetry request failed : error " + str(r.status_code))
            except:
                print("Telemetry request failed.")
        else:
            self.logger.info("Telemetry disabled, please consider enabling it by removing the 'telemetry=False' argument of your Device creation.")

        # We push our current state to make sure that
        # both our model and the hardware are synced.
        self._push_once()

    @property
    def services(self):
        return contList(self._services)

    @property
    def nodes(self):
        return nodeList(self._nodes)

    # Poll state from hardware.
    def _poll_once(self):
        self._state = self._io.read()
        if self._state != []:
            self._state['timestamp'] = time.time()
            return self._state
        return []

    def _poll_and_up(self):
        while self._running:
            if not self._pause :
                state = self._poll_once()
                if self._state != []:
                    self._update(state)
                    self._push_once()
            else :
                time.sleep(0.1)

    # Update our model with the new state.
    def _update(self, new_state):
        if 'dead_service' in new_state.keys() :
            #we have lost a service put a flag on this service
            alias = new_state['dead_service']
            if hasattr(self, alias):
                getattr(self, alias)._kill()
            if (self._freedomLink != None):
                self._freedomLink._kill(alias)
        if 'assert' in new_state.keys() :
            # A node assert, print assert informations
            if (('node_id' in new_state['assert']) and ('file' in new_state['assert']) and ('line' in new_state['assert'])):
                s = "************************* ASSERT *************************\n"
                s += "*  Node " + str(new_state['assert']['node_id']) + " assert in file " + new_state['assert']['file'] + " line " + str(new_state['assert']['line'])
                s += "\n**********************************************************"
                print (s)
            if (self._freedomLink != None):
                self._freedomLink._assert(alias)
        if 'services' not in new_state.keys():
            return

        for alias, mod in new_state['services'].items():
            if hasattr(self, alias):
                getattr(self, alias)._update(mod)
            if (self._freedomLink != None):
                self._freedomLink._update(alias, mod)

        self._last_update = time.time()

    def update_cmd(self, alias, key, val):
        with self._cmd_lock:
            self._cmd[alias][key] = val

    def update_data(self, alias, key, val, data):
        with self._cmd_lock:
            self._cmd_data.append({alias: {key: val}})
            self._binary.append(data.tobytes())

    def _push_once(self):
        with self._cmd_lock:
            if self._cmd:
                self._write( json.dumps({'services': self._cmd}).encode())
                self._cmd = defaultdict(lambda: defaultdict(lambda: None))
            for cmd, binary in zip(self._cmd_data, self._binary):
                time.sleep(0.01)
                self._write( json.dumps({'services': cmd}).encode() + '\n'.encode() + binary)

            self._cmd_data = []
            self._binary = []


    def _send(self, msg):
        with self._send_lock:
            self._io.send(msg)

    def _write(self, data):
        with self._send_lock:
            self._io.write(data)



from freedomrobotics.link import Link

import logging
import time

class FreedomLink(object):
    def __init__(self, device):
        self._link = Link("core", command_callback=self.callback)
        self._delegate = device

    def callback(msg):
        print("I receive:" + str(msg) )
        self._link.log("info", "I heard " + str(msg) )
        # Topic represent services.
        if hasattr(self._delegate, msg["topic"][1:]):
            service = getattr(self._delegate, msg["topic"][1:])
            print ("we have this service")
            # We have this service.
            if hasattr(service, msg["message"]):
                service_data = getattr(service, msg["message"])
                service_data = msg["message"][msg["message"]]

    def _update(self, alias, new_state):
        if 'io_state' in new_state:
            self._link.message("/" + alias + "/io_state", \
                    "sensor_msgs/Joy", \
                    {"io_state": new_state['io_state']})
        if 'temperature' in new_state:
            self._link.message("/" + alias + "/temperature", \
                    "sensor_msgs/Temperature", \
                    {"temperature": new_state['temperature']})
        if 'lux' in new_state:
            self._link.message("/" + alias + "/lux", \
                    "sensor_msgs/Illuminance", \
                    {"illuminance": new_state['lux']})
        if 'rot_position' in new_state:
            self._link.message("/" + alias + "/rot_position", \
                    "sensor_msgs/JointState", \
                    {"position": new_state['rot_position'], "name":alias})
        if 'trans_position' in new_state:
            self._link.message("/" + alias + "/trans_position", \
                    "sensor_msgs/Range", \
                    {"range": new_state['trans_position']})
        if 'rot_speed' in new_state:
            self._link.message("/" + alias + "/rot_speed", \
                    "sensor_msgs/JointState", \
                    {"velocity": new_state['rot_speed'], "name":alias})
        if 'trans_speed' in new_state:
            self._link.message("/" + alias + "/trans_speed", \
                    "sensor_msgs/JointState", \
                    {"velocity": new_state['trans_speed'], "name":alias})
        if 'force' in new_state:
            self._link.message("/" + alias + "/force", \
                    "sensor_msgs/JointState", \
                    {"effort": new_state['force'], "name":alias})
        if 'current' in new_state:
            self._link.message("/" + alias + "/current", \
                    "sensor_msgs/BatteryState", \
                    {"current": new_state['current']})
        if 'volt' in new_state:
            self._link.message("/" + alias + "/volt", \
                    "sensor_msgs/BatteryState", \
                    {"voltage": new_state['volt']})
        if 'quaternion' in new_state:
            self._link.message("/" + alias + "/quaternion", \
                    "geometry_msgs/Quaternion", \
                    {"y": new_state['quaternion'][0],"x": new_state['quaternion'][1],"z": new_state['quaternion'][2],"w": new_state['quaternion'][3]})
        if 'linear_accel' in new_state:
            self._link.message("/" + alias + "/linear_accel", \
                    "geometry_msgs/Accel", \
                    {"linear": {"y": new_state['linear_accel'][0],"x": new_state['linear_accel'][1],"z": new_state['linear_accel'][2]}})
        if 'accel' in new_state:
            self._link.message("/" + alias + "/accel", \
                    "geometry_msgs/Accel", \
                    {"angular": {"y": new_state['accel'][0],"x": new_state['accel'][1],"z": new_state['accel'][2]}})
        if 'gyro' in new_state:
            self._link.message("/" + alias + "/gyro", \
                    "geometry_msgs/Vector3", \
                    {"y": new_state['gyro'][0],"x": new_state['gyro'][1],"z": new_state['gyro'][2]})
        if 'euler' in new_state:
            self._link.message("/" + alias + "/euler", \
                    "geometry_msgs/Vector3", \
                    {"y": new_state['euler'][0],"x": new_state['euler'][1],"z": new_state['euler'][2]})
        if 'compass' in new_state:
            self._link.message("/" + alias + "/compass", \
                    "geometry_msgs/Vector3", \
                    {"y": new_state['compass'][0],"x": new_state['compass'][1],"z": new_state['compass'][2]})
        if 'gravity_vector' in new_state:
            self._link.message("/" + alias + "/gravity", \
                    "geometry_msgs/Vector3", \
                    {"y": new_state['gravity_vector'][0],"x": new_state['gravity_vector'][1],"z": new_state['gravity_vector'][2]})

# I don't know what to do with those ones :
        # if 'rotational_matrix' in new_state:
        #     self._rotational_matrix = new_state['rotational_matrix']
        # if 'pedometer' in new_state:
        #     self._pedometer = new_state['pedometer']
        # if 'walk_time' in new_state:
        #     self._walk_time = new_state['walk_time']

        # if 'heading' in new_state:
        #     self._heading = new_state['heading']
        # if 'revision' in new_state:
        #     self._firmware_revision = new_state['revision']
        # if 'luos_revision' in new_state:
        #     self._luos_revision = new_state['luos_revision']
        # if 'luos_statistics' in new_state:
        #     self._luos_statistics = new_state['luos_statistics']


    def _kill(self, alias):
        print ("service", alias, "have been excluded from the network due to no responses.")

    def _assert(self, alias):
        print ("service", alias, "assert.")
